"""
-------------
interface_DSP
-------------
Pythonic interface for getting/setting parameters of the DSPs
"""
# B. Swinkels, 2016

from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import object
from os import path

import logging
import math
import configparser

from virgotools.pvt_utils import add_property, check_frozen_setattr
from virgotools.cm_lib import cm_send_receive, CmError

try:
    import virgotools.pvt_internal_defaults as defaults
except ImportError:
    import virgotools.pvt_external_defaults as defaults

__all__ = ['DSP', 'DSPGetError', 'DSPSetError']

logger = logging.getLogger(__name__)


ALIAS = {'SIB1': 'IB', 'SDB1': 'OB'}
TIMEOUT = 3


class DSPGetError(Exception):
    pass


class DSPSetError(Exception):
    pass


class DSP(object):
    """
    convenience object for getting and setting values of 'gnames' of the
    suspension DSPs

    All gnames are available as attributes on the DSP object. Getting an attribute
    yields the current value, setting an attribute sets them with with a ramp-time equal
    to the ramptime attribute.

    Parameters
    ----------
    dsp_name : str
        name of the DSP, e.g. NE or SIB1
    ramptime : float
        ramptime in seconds

    Example
    -------
    >>> from virgotools import *
    >>> log_to_console()
    >>> WE = DSP('WE', ramptime=10)
    >>> WE.F0_X_SET
    0.0
    >>> WE.F0_X_SET = 1
    Ramping F0_X_SET attribute of satserver/we/1 in 10 seconds to 1
    >>> WE.F0_X_SET += 2
    Ramping F0_X_SET attribute of satserver/we/1 in 10 seconds to 3
    >>> WE.ramp_attribute('F0_X_SET', 0, 20)
    Ramping F0_X_SET attribute of satserver/we/1 in 20 seconds to 0
    """
    def __init__(self, dsp_name, ramptime=5.0):

        # replace aliases
        dsp_name = ALIAS.get(dsp_name, dsp_name)

        self.ramptime = ramptime  # private copy, not the Tango attribute!

        # always read the latest version of the ini file
        config = configparser.RawConfigParser()
        config.optionxform = str  # preserve case

        if not defaults.satserver_config_file:
            raise NameError("The default SatServer config is not defined, please set the 'satserver_config_file' variable in %s/pvt_external_defaults.py" %(path.dirname(path.abspath(__file__))))

        if not defaults.server:
            raise NameError("The default server is not defined, please set the 'server' variable in %s/pvt_external_defaults.py" %(path.dirname(path.abspath(__file__))))

        config.read(defaults.satserver_config_file)

        self._gname2crate = dict()

        crate_names = (dsp_name, "Sc_" + dsp_name, "Sa_" + dsp_name)
        found = False

        for crate in config.sections():
            crate = str(crate)  # FIXME remove when dropping Py27
            if crate in crate_names:
                found = True

                logger.debug('Reading gnames for crate %s from file %s', crate, defaults.satserver_config_file)

                for (gname, _val) in config.items(crate):

                    # check if top and bottom crates do not define same gname
                    if gname in self._gname2crate:
                        logger.error('Found same gname multiple times!')

                    self._gname2crate[gname] = crate

                    # see http://stackoverflow.com/a/2295372
                    prop = property(
                        lambda self, attr=gname:
                            self.get_attribute(attr),
                        lambda self, val, attr=gname: self.ramp_attribute(
                            attr, val, self.ramptime)
                    )
                    add_property(self, gname, prop)

        if not found:
            raise ValueError('Did not find any DSP crate with name %s, %s or %s' % crate_names)

        self._isfrozen = True  # protect against typos of attribute name

    def get_attribute(self, attr):
        """
        Parameters
        ----------
        attr : str
            name of gname (use tab-completion to see available ones)

        Returns
        -------
        value : float
            value of the gname

        Raises
        ------
        DSPGetError
            If something goes wrong with the Cm connection to the suspensions server, or
            if this server reports some error.
        """
        try:
            crate = self._gname2crate[attr]
        except KeyError:
            raise ValueError('unknown gname %r' % attr)

        try:
            reply = cm_send_receive('get_value_reply', TIMEOUT, defaults.server, 'get_value', crate, attr)
        except CmError as err:  # timeout or other
            raise DSPGetError("CmError while getting gname: " + str(err))

        if len(reply) != 1:
            raise DSPGetError("Got get_value_reply with unexpected payload: %r" % reply)
        value = reply[0]

        if not isinstance(value, float) or math.isnan(value) or math.isinf(value):
            raise DSPGetError('Did not receive a finite gname value: %r' % value)
        return value

    def ramp_attribute(self, attr, value, ramptime):
        """
        set attribute with explicit ramptime

        Parameters
        ----------
        attr : str
            name of Tango attribute (use tab-completion to see available ones)
        value : float
            new value for attribute
        ramptime : float
            time in seconds for changing the old value to the new value

        Raises
        ------
        DSPSetError
            If something goes wrong with the Cm connection to the suspensions server, or
            if this server reports some error.
        """
        try:
            crate = self._gname2crate[attr]
        except KeyError:
            raise ValueError('unknown gname %r' % attr)

        logger.info('Ramping %s attribute of crate %s in %g seconds to %g',
                    attr, crate, ramptime, value)
        if ramptime < 0:
            raise ValueError('Ramptime cannot be negative')
        if math.isnan(value) or math.isinf(value):
            raise ValueError('Value must be finite')

        try:
            reply = cm_send_receive(
                'set_value_reply', TIMEOUT, defaults.server, 'set_value', crate, attr,
                float(value), float(ramptime))
        except CmError as err:  # timeout or other
            raise DSPSetError("CmError while setting gname: " + str(err))

        if not (len(reply) == 1 and reply[0] == "OK"):
            logger.error('Received unexpected set_value_reply: %r' % reply)
            raise DSPSetError("Failure in setting gname: %r" % reply)

    def slew_attribute(self, attr, value, slewrate):
        """
        set attribute with a given slewrate

        Parameters
        ----------
        attr : str
            name of Tango attribute
        value : float
            new value for attribute
        slewrate : float
            maximum slewrate in units per second (e.g. urad/sec for alignment)

        Returns
        -------
        ramptime : float
            time in seconds to perform the ramp
        """
        if slewrate <= 0:
            raise ValueError('slewrate must be positive')

        ramptime = abs(value - self.get_attribute(attr)) / slewrate
        if ramptime > 3600:
            raise ValueError('ramp would take too long')

        self.ramp_attribute(attr, value, ramptime)

        return ramptime

    __setattr__ = check_frozen_setattr()
