"""
some utilities for internal use in PythonVirgoTools
"""
# B. Swinkels, 2016


# http://stackoverflow.com/a/14620633
class AttrDict(dict):
    """Dictionary that allows lookup of items using attribute access"""
    _repr_level = 0  # not thread safe

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def __repr__(self):
        # print nested AttrDicts in a pretty way to make VPM more readable
        self.__class__._repr_level += 1
        try:
            if self._repr_level == 1:
                items = []
                for k, v in self.items():
                    if isinstance(v, AttrDict):
                        items.append('  {}={},\n'.format(k, repr(v)))
                    else:
                        self.__class__._repr_level += 1
                        items.append('  {}={},\n'.format(k, repr(v)))
                        self.__class__._repr_level -= 1
                return '{}(\n{})'.format(self.__class__.__name__,
                                         ''.join(items))
            elif self._repr_level == 2:
                items = []
                for k, v in self.items():
                    if isinstance(v, AttrDict):
                        rv = repr(v)
                    elif isinstance(v, list):
                        rv = '[...]'
                    elif isinstance(v, (dict, set)):
                        rv = '{...}'
                    else:
                        rv = repr(v)
                    items.append('{}: {}'.format(repr(k), rv))
                return '{{{}}}'.format(', '.join(items))
            else:
                return '<{}>'.format(getattr(self, 'name', '...'))
        finally:
            self.__class__._repr_level -= 1


# dynamically add property to instance,
# see http://stackoverflow.com/a/2954373 and comments
def add_property(inst, name, prop):
    """dynamically add property to instance"""
    cls = type(inst)
    if not hasattr(cls, '__perinstance'):
        cls = type(cls.__name__, (cls,), {})
        cls.__perinstance = True
        inst.__class__ = cls
    setattr(cls, name, prop)


def check_frozen_setattr(msg=None):
    """
    Alternative __setattr__ method for preventing addition of new attributes.

    Example
    -------
    class C(object):
        def __init__(self, val1, val2):
            self._val1 = val1
            self._val2 = val2
            self._isfrozen = True

        __setattr__ = check_frozen_setattr("adding new attr {} is forbidden!")

    """
    if msg is None:
        msg = "Object if frozen, cannot add new attribute {}"

    # http://stackoverflow.com/a/3603824
    # changed hasattr(self, attr) by attr in dir(self) to avoid trouble when
    # getter raises an error, see https://hynek.me/articles/hasattr/
    def __setattr__(self, attr, value):
        if getattr(self, '_isfrozen', False) and attr not in dir(self):
            raise AttributeError(msg.format(attr))
        object.__setattr__(self, attr, value)

    return __setattr__


# FIXME: change this to e.g. ascii_str, since there is already a new function in
# Python3 called ascii
def ascii(s):
    """converts bytes, unicode or ctypesgen strings to a byte string on both Python2 and
    Python3 (so to str in Python 2 and bytes in Python3). Raises an error for unicode
    strings containing characters outside of ascii"""

    # if s is a ctypesgen wrapper for a string, get the payload
    s = getattr(s, 'data', s)

    if isinstance(s, bytes):
        return s
    return s.encode('ascii', 'strict')


def native_str(s):
    """converts bytes, unicode or ctypesgen strings to a native str on both Python2 and
    Python3 (so a byte string on Python2 and a unicode string on Python3). Raises an
    error on Python2 for unicode strings containing characters outside of ascii, or when
    bytes cannot be decoded to unicode on Python3"""

    # if s is a ctypesgen wrapper for a string, get the payload
    s = getattr(s, 'data', s)

    if str is bytes:
        # Python 2
        if isinstance(s, bytes):
            return s
        return s.encode('ascii', 'strict')

    # Python 3
    if isinstance(s, str):
        return s
    return s.decode('ascii', 'strict')
